@@ -14,7 +14,7 @@ copyright = '2020, Stuart Mumford and Contributors'
 author = 'Stuart Mumford'
 
 # The full version, including alpha/beta/rc tags
-from parfive import __version__ as release
+from parfive import __version__ as release  # NOQA
 
 # -- General configuration ---------------------------------------------------
 
@@ -50,6 +50,10 @@ source_suffix = '.rst'
 # The master toctree document.
 master_doc = 'index'
 
+# The reST default role (used for this markup: `text`) to use for all
+# documents. Set to the "smart" one.
+default_role = 'obj'
+
 # -- Options for intersphinx extension ---------------------------------------
 
 # Example configuration for intersphinx: refer to the Python standard library.
@@ -60,7 +64,7 @@ intersphinx_mapping = {'https://docs.python.org/': None,
 # -- Options for HTML output -------------------------------------------------
 
 try:
-    from sunpy_sphinx_theme.conf import *
+    from sunpy_sphinx_theme.conf import *  # NOQA
 except ImportError:
     html_theme = 'alabaster'
 
