@@ -1,24 +1,33 @@
-import asyncio
-import contextlib
 import os
 import sys
-
+import asyncio
 import pathlib
+import warnings
+import contextlib
 import urllib.parse
-from concurrent.futures import ThreadPoolExecutor
 from functools import partial
+from concurrent.futures import ThreadPoolExecutor
 
-import parfive
 import aiohttp
 from tqdm import tqdm, tqdm_notebook
 
+import parfive
 from .results import Results
-from .utils import (FailedDownload, Token, default_name, get_filepath,
-                    get_ftp_size, get_http_size, in_notebook, run_in_thread)
+from .utils import (
+    FailedDownload,
+    Token,
+    _QueueList,
+    default_name,
+    get_filepath,
+    get_ftp_size,
+    get_http_size,
+    in_notebook,
+    run_in_thread,
+)
 
 try:
     import aioftp
-except ImportError:
+except ImportError:  # pragma: nocover
     aioftp = None
 
 
@@ -31,45 +40,38 @@ class Downloader:
 
     Parameters
     ----------
-
     max_conn : `int`, optional
         The number of parallel download slots.
-
     progress : `bool`, optional
         If `True` show a main progress bar showing how many of the total files
         have been downloaded. If `False`, no progress bars will be shown at all.
-
     file_progress : `bool`, optional
         If `True` and ``progress`` is true, show ``max_conn`` progress bars
         detailing the progress of each individual file being downloaded.
-
     loop : `asyncio.AbstractEventLoop`, optional
-        The event loop to use to download the files. If not specified a new
-        loop will be created and executed in a new thread so it does not
-        interfere with any currently running event loop.
-
+        No longer used, and will be removed in a future release.
     notebook : `bool`, optional
         If `True` tqdm will be used in notebook mode. If `None` an attempt will
         be made to detect the notebook and guess which progress bar to use.
-
     overwrite : `bool` or `str`, optional
         Determine how to handle downloading if a file already exists with the
         same name. If `False` the file download will be skipped and the path
         returned to the existing file, if `True` the file will be downloaded
         and the existing file will be overwritten, if `'unique'` the filename
         will be modified to be unique.
-
     headers : `dict`
        Request headers to be passed to the server.
        Adds `User-Agent` information about `parfive`, `aiohttp` and `python` if not passed explicitely.
-
     """
 
     def __init__(self, max_conn=5, progress=True, file_progress=True,
                  loop=None, notebook=None, overwrite=False, headers=None):
 
+        if loop:
+            warnings.warn('The loop argument is no longer used, and will be '
+                          'removed in a future release.')
         self.max_conn = max_conn
-        self._start_loop(loop)
+        self._init_queues()
 
         # Configure progress bars
         if notebook is None:
@@ -82,28 +84,20 @@ class Downloader:
 
         self.headers = headers
         if headers is None or 'User-Agent' not in headers:
-            self.headers = {'User-Agent': f"parfive/{parfive.__version__} aiohttp/{aiohttp.__version__} python/{sys.version[:5]}"}
-
-    def _start_loop(self, loop):
-        # Setup asyncio loops
-        if not loop:
-            aio_pool = ThreadPoolExecutor(1)
-            self.loop = asyncio.new_event_loop()
-            self.run_until_complete = partial(run_in_thread, aio_pool, self.loop)
-        else:
-            self.loop = loop
-            self.run_until_complete = self.loop.run_until_complete
-
-        asyncio.set_event_loop(self.loop)
+            self.headers = {
+                'User-Agent': f"parfive/{parfive.__version__} aiohttp/{aiohttp.__version__} python/{sys.version[:5]}"}
 
+    def _init_queues(self):
         # Setup queues
-        self.http_queue = asyncio.Queue(loop=self.loop)
-        self.http_tokens = asyncio.Queue(maxsize=self.max_conn, loop=self.loop)
-        self.ftp_queue = asyncio.Queue(loop=self.loop)
-        self.ftp_tokens = asyncio.Queue(maxsize=self.max_conn, loop=self.loop)
+        self.http_queue = _QueueList()
+        self.ftp_queue = _QueueList()
+
+    def _generate_tokens(self):
+        # Create a Queue with max_conn tokens
+        queue = asyncio.Queue(maxsize=self.max_conn)
         for i in range(self.max_conn):
-            self.http_tokens.put_nowait(Token(i + 1))
-            self.ftp_tokens.put_nowait(Token(i + 1))
+            queue.put_nowait(Token(i + 1))
+        return queue
 
     @property
     def queued_downloads(self):
@@ -111,7 +105,7 @@ class Downloader:
         The total number of files already queued for download.
         """
 
-        return self.http_queue.qsize() + self.ftp_queue.qsize()
+        return len(self.http_queue) + len(self.ftp_queue)
 
     def enqueue_file(self, url, path=None, filename=None, overwrite=None, **kwargs):
         """
@@ -121,18 +115,15 @@ class Downloader:
         ----------
         url : `str`
             The URL to retrieve.
-
         path : `str`, optional
             The directory to retrieve the file into, if `None` defaults to the
             current directory.
-
         filename : `str` or `callable`, optional
             The filename to save the file as. Can also be a callable which
             takes two arguments the url and the response object from opening
             that URL, and returns the filename. (Note, for FTP downloads the
             response will be ``None``.) If `None` the HTTP headers will be read
             for the filename, or the last segment of the URL will be used.
-
         overwrite : `bool` or `str`, optional
             Determine how to handle downloading if a file already exists with the
             same name. If `False` the file download will be skipped and the path
@@ -140,7 +131,6 @@ class Downloader:
             and the existing file will be overwritten, if `'unique'` the filename
             will be modified to be unique. If `None` the value set when
             constructing the `~parfive.Downloader` object will be used.
-
         kwargs : `dict`
             Extra keyword arguments are passed to `aiohttp.ClientSession.get`
             or `aioftp.Client.context` depending on the protocol.
@@ -152,7 +142,6 @@ class Downloader:
         Authentication `proxy_auth` should be passed as a `aiohttp.BasicAuth`
         object. Proxy Headers `proxy_headers` should be passed as `dict`
         object.
-
         """
         overwrite = overwrite or self.overwrite
 
@@ -176,16 +165,86 @@ class Downloader:
         if scheme in ('http', 'https'):
             get_file = partial(self._get_http, url=url, filepath_partial=filepath,
                                overwrite=overwrite, **kwargs)
-            self.http_queue.put_nowait(get_file)
+            self.http_queue.append(get_file)
         elif scheme == 'ftp':
             if aioftp is None:
                 raise ValueError("The aioftp package must be installed to download over FTP.")
             get_file = partial(self._get_ftp, url=url, filepath_partial=filepath,
                                overwrite=overwrite, **kwargs)
-            self.ftp_queue.put_nowait(get_file)
+            self.ftp_queue.append(get_file)
         else:
             raise ValueError("URL must start with either 'http' or 'ftp'.")
 
+    @staticmethod
+    def _run_in_loop(coro):
+        """
+        Detect an existing, running loop and run in a separate loop if needed.
+
+        If no loop is running, use asyncio.run to run the coroutine instead.
+        """
+        try:
+            loop = asyncio.get_running_loop()
+        except RuntimeError:
+            loop = None
+
+        if loop and loop.is_running():
+            aio_pool = ThreadPoolExecutor(1)
+            new_loop = asyncio.new_event_loop()
+            return run_in_thread(aio_pool, new_loop, coro)
+
+        return asyncio.run(coro)
+
+    async def run_download(self, timeouts=None):
+        """
+        Download all files in the queue.
+
+        Parameters
+        ----------
+        timeouts : `dict`, optional
+            Overrides for the default timeouts for http downloads. Supported
+            keys are any accepted by the `aiohttp.ClientTimeout` class. Defaults
+            to 5 minutes for total session timeout and 90 seconds for socket
+            read timeout.
+
+        Returns
+        -------
+        `parfive.Results`
+            A list of files downloaded.
+
+        Notes
+        -----
+        The defaults for the `'total'` and `'sock_read'` timeouts can be
+        overridden by two environment variables ``PARFIVE_TOTAL_TIMEOUT`` and
+        ``PARFIVE_SOCK_READ_TIMEOUT``.
+        """
+        timeouts = timeouts or {"total": os.environ.get("PARFIVE_TOTAL_TIMEOUT", 5 * 60),
+                                "sock_read": os.environ.get("PARFIVE_SOCK_READ_TIMEOUT", 90)}
+
+        total_files = self.queued_downloads
+
+        done = set()
+        with self._get_main_pb(total_files) as main_pb:
+            if len(self.http_queue):
+                done.update(await self._run_http_download(main_pb, timeouts))
+            if len(self.ftp_queue):
+                done.update(await self._run_ftp_download(main_pb, timeouts))
+
+        dl_results = await asyncio.gather(*done, return_exceptions=True)
+
+        results = Results()
+
+        # Iterate through the results and store any failed download errors in
+        # the errors list of the results object.
+        for res in dl_results:
+            if isinstance(res, FailedDownload):
+                results.add_error(res.filepath_partial, res.url, res.exception)
+            elif isinstance(res, Exception):
+                raise res
+            else:
+                results.append(res)
+
+        return results
+
     def download(self, timeouts=None):
         """
         Download all files in the queue.
@@ -200,37 +259,20 @@ class Downloader:
 
         Returns
         -------
-        filenames : `parfive.Results`
+        `parfive.Results`
             A list of files downloaded.
 
         Notes
         -----
+        This is a synchronous version of `~parfive.Downloader.run_download`, an
+        `asyncio` event loop will be created to run the download (in it's own
+        thread if a loop is already running).
+
         The defaults for the `'total'` and `'sock_read'` timeouts can be
         overridden by two environment variables ``PARFIVE_TOTAL_TIMEOUT`` and
         ``PARFIVE_SOCK_READ_TIMEOUT``.
-
         """
-        timeouts = timeouts or {"total": os.environ.get("PARFIVE_TOTAL_TIMEOUT", 5 * 60),
-                                "sock_read": os.environ.get("PARFIVE_SOCK_READ_TIMEOUT", 90)}
-        try:
-            future = self.run_until_complete(self._run_download(timeouts))
-        finally:
-            self.loop.stop()
-        dlresults = future.result()
-
-        results = Results()
-
-        # Iterate through the results and store any failed download errors in
-        # the errors list of the results object.
-        for res in dlresults:
-            if isinstance(res, FailedDownload):
-                results.add_error(res.filepath_partial, res.url, res.exception)
-            elif isinstance(res, Exception):
-                raise res
-            else:
-                results.append(res)
-
-        return results
+        return self._run_in_loop(self.run_download(timeouts))
 
     def retry(self, results):
         """
@@ -241,22 +283,19 @@ class Downloader:
 
         Parameters
         ----------
-
         results : `parfive.Results`
             A previous results object, the ``.errors`` property will be read
             and the downloads retried.
 
         Returns
         -------
-
-        results : `parfive.Results`
+        `parfive.Results`
             A modified version of the input ``results`` with all the errors from
             this download attempt and any new files appended to the list of
             file paths.
-
         """
-        # Restart the loop.
-        self._start_loop(None)
+        # Reset the queues
+        self._init_queues()
 
         for err in results.errors:
             self.enqueue_file(err.url, filename=err.filepath_partial)
@@ -280,44 +319,26 @@ class Downloader:
         else:
             return contextlib.contextmanager(lambda: iter([None]))()
 
-    async def _run_download(self, timeouts):
-        """
-        Download all files in the queue.
-
-        Returns
-        -------
-
-        results : `parfive.Results`
-            A list of filenames which successfully downloaded. This list also
-            has an attribute ``errors`` which lists any failed urls and their
-            error.
-        """
-        total_files = self.http_queue.qsize() + self.ftp_queue.qsize()
-        done = set()
-        with self._get_main_pb(total_files) as main_pb:
-            if not self.http_queue.empty():
-                done.update(await self._run_http_download(main_pb, timeouts))
-            if not self.ftp_queue.empty():
-                done.update(await self._run_ftp_download(main_pb, timeouts))
-
-        # Return one future to represent all the results.
-        return asyncio.gather(*done, return_exceptions=True)
-
     async def _run_http_download(self, main_pb, timeouts):
         async with aiohttp.ClientSession(headers=self.headers) as session:
-            futures = await self._run_from_queue(self.http_queue, self.http_tokens,
-                                                 main_pb, session=session, timeouts=timeouts)
+            self._generate_tokens()
+            futures = await self._run_from_queue(
+                self.http_queue.generate_queue(),
+                self._generate_tokens(),
+                main_pb, session=session, timeouts=timeouts)
 
             # Wait for all the coroutines to finish
-            done, _ = await asyncio.wait(futures, loop=self.loop)
+            done, _ = await asyncio.wait(futures)
 
             return done
 
     async def _run_ftp_download(self, main_pb, timeouts):
-        futures = await self._run_from_queue(self.ftp_queue, self.ftp_tokens,
-                                             main_pb, timeouts=timeouts)
+        futures = await self._run_from_queue(
+            self.ftp_queue.generate_queue(),
+            self._generate_tokens(),
+            main_pb, timeouts=timeouts)
         # Wait for all the coroutines to finish
-        done, _ = await asyncio.wait(futures, loop=self.loop)
+        done, _ = await asyncio.wait(futures)
 
         return done
 
@@ -328,8 +349,8 @@ class Downloader:
             token = await tokens.get()
             file_pb = self.tqdm if self.file_progress else False
             future = asyncio.ensure_future(get_file(session, token=token,
-                                                    file_pb=file_pb, timeouts=timeouts),
-                                           loop=self.loop)
+                                                    file_pb=file_pb,
+                                                    timeouts=timeouts))
 
             def callback(token, future, main_pb):
                 tokens.put_nowait(token)
@@ -349,38 +370,28 @@ class Downloader:
 
         Parameters
         ----------
-
         session : `aiohttp.ClientSession`
             The `aiohttp.ClientSession` to use to retrieve the files.
-
         url : `str`
             The url to retrieve.
-
         filepath_partial : `callable`
             A function to call which returns the filepath to save the url to.
             Takes two arguments ``resp, url``.
-
         chunksize : `int`
             The number of bytes to read into the file at a time.
-
         file_pb : `tqdm.tqdm` or `False`
             Should progress bars be displayed for each file downloaded.
-
         token : `parfive.downloader.Token`
             A token for this download slot.
-
         max_splits: `int`
             Number of maximum concurrent connections per file.
-
         kwargs : `dict`
             Extra keyword arguments are passed to `aiohttp.ClientSession.get`.
 
         Returns
         -------
-
-        filepath : `str`
+        `str`
             The name of the file saved.
-
         """
         timeout = aiohttp.ClientTimeout(**timeouts)
         try:
@@ -409,7 +420,7 @@ class Downloader:
                     downloaded_chunk_queue = asyncio.Queue()
 
                     download_workers = []
-                    writer = self.loop.create_task(
+                    writer = asyncio.create_task(
                         self._write_worker(downloaded_chunk_queue, file_pb, filepath))
 
                     if max_splits and resp.headers.get('Accept-Ranges', None) == "bytes":
@@ -423,13 +434,13 @@ class Downloader:
                         ranges[-1][1] = ''
                         for _range in ranges:
                             download_workers.append(
-                                self.loop.create_task(self._http_download_worker(
+                                asyncio.create_task(self._http_download_worker(
                                     session, url, chunksize, _range, timeout, downloaded_chunk_queue, **kwargs
                                 ))
                             )
                     else:
                         download_workers.append(
-                            self.loop.create_task(self._http_download_worker(
+                            asyncio.create_task(self._http_download_worker(
                                 session, url, chunksize, None, timeout, downloaded_chunk_queue, **kwargs
                             ))
                         )
@@ -454,13 +465,11 @@ class Downloader:
 
         Parameters
         ----------
-
         queue: `asyncio.Queue`
              Queue for chunks
 
         file_pb : `tqdm.tqdm` or `False`
             Should progress bars be displayed for each file downloaded.
-
         filepath: `pathlib.Path`
             Path to the which the file should be downloaded.
         """
@@ -488,23 +497,17 @@ class Downloader:
 
         Parameters
         ----------
-
         session : `aiohttp.ClientSession`
             The `aiohttp.ClientSession` to use to retrieve the files.
-
         url : `str`
             The url to retrieve.
-
         chunksize : `int`
             The number of bytes to read into the file at a time.
-
         http_range: (`int`, `int`) or `None`
             Start and end bytes of the file. In None, then no `Range` header is specified
             in request and the whole file will be downloaded.
-
         queue: `asyncio.Queue`
              Queue to put the download chunks.
-
         kwargs : `dict`
             Extra keyword arguments are passed to `aiohttp.ClientSession.get`.
         """
@@ -531,32 +534,24 @@ class Downloader:
 
         Parameters
         ----------
-
         session : `None`
             A placeholder for API compatibility with ``_get_http``
-
         url : `str`
             The url to retrieve.
-
         filepath_partial : `callable`
             A function to call which returns the filepath to save the url to.
             Takes two arguments ``resp, url``.
-
         file_pb : `tqdm.tqdm` or `False`
             Should progress bars be displayed for each file downloaded.
-
         token : `parfive.downloader.Token`
             A token for this download slot.
-
         kwargs : `dict`
             Extra keyword arguments are passed to `~aioftp.Client.context`.
 
         Returns
         -------
-
-        filepath : `str`
+        `str`
             The name of the file saved.
-
         """
         parse = urllib.parse.urlparse(url)
         try:
@@ -578,11 +573,12 @@ class Downloader:
 
                     downloaded_chunks_queue = asyncio.Queue()
                     download_workers = []
-                    writer = self.loop.create_task(
+                    writer = asyncio.create_task(
                         self._write_worker(downloaded_chunks_queue, file_pb, filepath))
 
                     download_workers.append(
-                        self.loop.create_task(self._ftp_download_worker(stream, downloaded_chunks_queue))
+                        asyncio.create_task(self._ftp_download_worker(
+                            stream, downloaded_chunks_queue))
                     )
 
                     await asyncio.gather(*download_workers)
@@ -596,11 +592,11 @@ class Downloader:
 
     async def _ftp_download_worker(self, stream, queue):
         """
-        Similar to `Downloader._http_download_worker`. See that function's documentation for more info.
+        Similar to `Downloader._http_download_worker`.
+        See that function's documentation for more info.
 
         Parameters
         ----------
-
         stream: `aioftp.StreamIO`
             Stream of the file to be downloaded.
         queue: `asyncio.Queue`
