@@ -1,72 +1,97 @@
+import os
+import sys
 import platform
 from pathlib import Path
 from unittest import mock
 from unittest.mock import patch
 
 import aiohttp
-from aiohttp import ClientTimeout
-from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
-
 import pytest
+from aiohttp import ClientTimeout
 from pytest_localserver.http import WSGIServer
-from parfive.downloader import Downloader, Token, FailedDownload, Results
-from parfive.utils import sha256sum
+
 import parfive
-import sys
-import os
-import hashlib
+from parfive.downloader import Downloader, FailedDownload, Results, Token
+from parfive.utils import sha256sum
 
 skip_windows = pytest.mark.skipif(platform.system() == 'Windows', reason="Windows.")
 
-def test_setup(event_loop):
-    dl = Downloader(loop=event_loop)
+
+def validate_test_file(f):
+    assert len(f) == 1
+    assert Path(f[0]).name == "testfile.fits"
+    assert sha256sum(f[0]) == "a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3"
+
+
+def test_setup():
+    dl = Downloader()
 
     assert isinstance(dl, Downloader)
 
-    assert dl.http_queue.qsize() == 0
-    assert dl.http_tokens.qsize() == 5
-    assert dl.ftp_queue.qsize() == 0
-    assert dl.ftp_tokens.qsize() == 5
+    assert len(dl.http_queue) == 0
+    assert len(dl.ftp_queue) == 0
+    assert dl._generate_tokens().qsize() == 5
 
 
-def test_download(event_loop, httpserver, tmpdir):
+def test_download(httpserver, tmpdir):
     tmpdir = str(tmpdir)
     httpserver.serve_content('SIMPLE  = T',
                              headers={'Content-Disposition': "attachment; filename=testfile.fits"})
-    dl = Downloader(loop=event_loop)
+    dl = Downloader()
 
     dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None)
 
     assert dl.queued_downloads == 1
 
     f = dl.download()
-
-    assert len(f) == 1
-    assert Path(f[0]).name == "testfile.fits"
-    assert sha256sum(f[0]) == "a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3"
+    validate_test_file(f)
 
 
-def test_download_ranged_http(event_loop, httpserver, tmpdir):
+def test_changed_max_conn(httpserver, tmpdir):
+    # Check that changing max_conn works after creating Downloader
     tmpdir = str(tmpdir)
     httpserver.serve_content('SIMPLE  = T',
                              headers={'Content-Disposition': "attachment; filename=testfile.fits"})
-    dl = Downloader(loop=event_loop)
+    dl = Downloader(max_conn=4)
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None)
+    dl.max_conn = 3
+
+    f = dl.download()
+    validate_test_file(f)
+
+
+@pytest.mark.asyncio
+async def test_async_download(httpserver, tmpdir):
+    httpserver.serve_content('SIMPLE  = T',
+                             headers={'Content-Disposition': "attachment; filename=testfile.fits"})
+    dl = Downloader()
+
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None)
+
+    assert dl.queued_downloads == 1
+
+    f = await dl.run_download()
+    validate_test_file(f)
+
+
+def test_download_ranged_http(httpserver, tmpdir):
+    tmpdir = str(tmpdir)
+    httpserver.serve_content('SIMPLE  = T',
+                             headers={'Content-Disposition': "attachment; filename=testfile.fits"})
+    dl = Downloader()
 
     dl.enqueue_file(httpserver.url, path=Path(tmpdir))
 
     assert dl.queued_downloads == 1
 
     f = dl.download()
-
-    assert len(f) == 1
-    assert Path(f[0]).name == "testfile.fits"
-    assert sha256sum(f[0]) == "a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3"
+    validate_test_file(f)
 
 
-def test_download_partial(event_loop, httpserver, tmpdir):
+def test_download_partial(httpserver, tmpdir):
     tmpdir = str(tmpdir)
     httpserver.serve_content('SIMPLE  = T')
-    dl = Downloader(loop=event_loop)
+    dl = Downloader()
 
     dl.enqueue_file(httpserver.url, filename=lambda resp, url: Path(tmpdir) / "filename")
     f = dl.download()
@@ -76,14 +101,14 @@ def test_download_partial(event_loop, httpserver, tmpdir):
     assert "filename" in f[0]
 
 
-def test_empty_download(event_loop, tmpdir):
-    dl = Downloader(loop=event_loop)
+def test_empty_download(tmpdir):
+    dl = Downloader()
 
     f = dl.download()
     assert len(f) == 0
 
 
-def test_download_filename(event_loop, httpserver, tmpdir):
+def test_download_filename(httpserver, tmpdir):
     httpserver.serve_content('SIMPLE  = T')
 
     fname = "testing123"
@@ -91,7 +116,7 @@ def test_download_filename(event_loop, httpserver, tmpdir):
     with open(filename, "w") as fh:
         fh.write("SIMPLE = T")
 
-    dl = Downloader(loop=event_loop)
+    dl = Downloader()
 
     dl.enqueue_file(httpserver.url, filename=filename, chunksize=200)
     f = dl.download()
@@ -102,7 +127,7 @@ def test_download_filename(event_loop, httpserver, tmpdir):
     assert f[0] == filename
 
 
-def test_download_no_overwrite(event_loop, httpserver, tmpdir):
+def test_download_no_overwrite(httpserver, tmpdir):
     httpserver.serve_content('SIMPLE  = T')
 
     fname = "testing123"
@@ -110,7 +135,7 @@ def test_download_no_overwrite(event_loop, httpserver, tmpdir):
     with open(filename, "w") as fh:
         fh.write("Hello world")
 
-    dl = Downloader(loop=event_loop)
+    dl = Downloader()
 
     dl.enqueue_file(httpserver.url, filename=filename, chunksize=200)
     f = dl.download()
@@ -126,7 +151,7 @@ def test_download_no_overwrite(event_loop, httpserver, tmpdir):
         assert fh.read() == "Hello world"
 
 
-def test_download_overwrite(event_loop, httpserver, tmpdir):
+def test_download_overwrite(httpserver, tmpdir):
     httpserver.serve_content('SIMPLE  = T')
 
     fname = "testing123"
@@ -134,7 +159,7 @@ def test_download_overwrite(event_loop, httpserver, tmpdir):
     with open(filename, "w") as fh:
         fh.write("Hello world")
 
-    dl = Downloader(loop=event_loop, overwrite=True)
+    dl = Downloader(overwrite=True)
 
     dl.enqueue_file(httpserver.url, filename=filename, chunksize=200)
     f = dl.download()
@@ -148,15 +173,15 @@ def test_download_overwrite(event_loop, httpserver, tmpdir):
         assert fh.read() == "SIMPLE  = T"
 
 
-def test_download_unique(event_loop, httpserver, tmpdir):
+def test_download_unique(httpserver, tmpdir):
     httpserver.serve_content('SIMPLE  = T')
 
     fname = "testing123"
     filename = str(tmpdir.join(fname))
 
-    filenames = [filename, filename+'.fits', filename+'.fits.gz']
+    filenames = [filename, filename + '.fits', filename + '.fits.gz']
 
-    dl = Downloader(loop=event_loop, overwrite='unique')
+    dl = Downloader(overwrite='unique')
 
     # Write files to both the target filenames.
     for fn in filenames:
@@ -177,11 +202,15 @@ def test_download_unique(event_loop, httpserver, tmpdir):
 
 @pytest.fixture
 def testserver(request):
-    """A server that throws a 404 for the second request"""
+    """
+    A server that throws a 404 for the second request.
+    """
     counter = 0
 
     def simple_app(environ, start_response):
-        """Simplest possible WSGI application"""
+        """
+        Simplest possible WSGI application.
+        """
         nonlocal counter
 
         counter += 1
@@ -347,7 +376,6 @@ def test_ftp_http(tmpdir, httpserver):
     dl.enqueue_file(httpserver.url, path=tmpdir)
     dl.enqueue_file("http://noaurl.notadomain/noafile", path=tmpdir)
 
-
     assert dl.queued_downloads == 6
 
     f = dl.download()
@@ -355,12 +383,12 @@ def test_ftp_http(tmpdir, httpserver):
     assert len(f.errors) == 4
 
 
-def test_default_user_agent(event_loop, httpserver, tmpdir):
+def test_default_user_agent(httpserver, tmpdir):
     tmpdir = str(tmpdir)
     httpserver.serve_content('SIMPLE  = T',
                              headers={'Content-Disposition': "attachment; filename=testfile.fits"})
 
-    dl = Downloader(loop=event_loop)
+    dl = Downloader()
     dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None)
 
     assert dl.queued_downloads == 1
@@ -368,15 +396,16 @@ def test_default_user_agent(event_loop, httpserver, tmpdir):
     dl.download()
 
     assert 'User-Agent' in httpserver.requests[0].headers
-    assert httpserver.requests[0].headers['User-Agent'] == f"parfive/{parfive.__version__} aiohttp/{aiohttp.__version__} python/{sys.version[:5]}"
+    assert httpserver.requests[0].headers[
+        'User-Agent'] == f"parfive/{parfive.__version__} aiohttp/{aiohttp.__version__} python/{sys.version[:5]}"
 
 
-def test_custom_user_agent(event_loop, httpserver, tmpdir):
+def test_custom_user_agent(httpserver, tmpdir):
     tmpdir = str(tmpdir)
     httpserver.serve_content('SIMPLE  = T',
                              headers={'Content-Disposition': "attachment; filename=testfile.fits"})
 
-    dl = Downloader(loop=event_loop, headers={'User-Agent': 'test value 299792458'})
+    dl = Downloader(headers={'User-Agent': 'test value 299792458'})
     dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None)
 
     assert dl.queued_downloads == 1
@@ -387,16 +416,17 @@ def test_custom_user_agent(event_loop, httpserver, tmpdir):
     assert httpserver.requests[0].headers['User-Agent'] == "test value 299792458"
 
 
-@patch.dict(os.environ, {'HTTP_PROXY': "http_proxy_url",'HTTPS_PROXY': "https_proxy_url"})
-@pytest.mark.parametrize("url,proxy",[('http://test.example.com','http_proxy_url'),('https://test.example.com','https_proxy_url')])
-def test_proxy_passed_as_kwargs_to_get(event_loop, tmpdir, url, proxy):
+@patch.dict(os.environ, {'HTTP_PROXY': "http_proxy_url", 'HTTPS_PROXY': "https_proxy_url"})
+@pytest.mark.parametrize("url,proxy", [('http://test.example.com',
+                                        'http_proxy_url'), ('https://test.example.com', 'https_proxy_url')])
+def test_proxy_passed_as_kwargs_to_get(tmpdir, url, proxy):
 
     with mock.patch(
-                    "aiohttp.client.ClientSession._request",
-                    new_callable=mock.MagicMock
-                   ) as patched:
+        "aiohttp.client.ClientSession._request",
+        new_callable=mock.MagicMock
+    ) as patched:
 
-        dl = Downloader(loop=event_loop)
+        dl = Downloader()
         dl.enqueue_file(url, path=Path(tmpdir), max_splits=None)
 
         assert dl.queued_downloads == 1
@@ -405,7 +435,7 @@ def test_proxy_passed_as_kwargs_to_get(event_loop, tmpdir, url, proxy):
 
     assert patched.called, "`ClientSession._request` not called"
     assert list(patched.call_args) == [('GET', url),
-                                       {'allow_redirects': True, 
+                                       {'allow_redirects': True,
                                         'timeout': ClientTimeout(total=300, connect=None, sock_read=90, sock_connect=None),
                                         'proxy': proxy
-                                       }]
+                                        }]
